import datetime
import discord
from contextlib import redirect_stdout
from discord.ext import commands
from io import StringIO
from textwrap import indent
from traceback import format_exc

class CogConverter(commands.Converter):
    async def convert(self, ctx: commands.Context, argument: str) -> commands.Cog:
        cog = ctx.bot.get_cog(argument)
        if cog is None:
            raise commands.ExtensionNotFound(argument)
        return cog


class Meta(commands.Cog):
    """
    Default commands and listeners and stuffs
    """

    def __init__(self, bot):
        self.bot = bot
    
    @staticmethod
    def _cleanup_code(content: str) -> str:
        """Automatically removes code blocks from the code.

        This is just used for the `execute` command at this time
        
        Parameters
        ----------
        content : str
            The content to remove code blocks from
        
        Returns
        -------
        The cleaned content as a string
        """

        # remove \`\`\`py\n\`\`\`
        if content.startswith('```') and content.endswith('```'):
            if content[-4] == '\n':
                return '\n'.join(content.split('\n')[1:-1])
            return '\n'.join(content.split('\n')[1:]).rstrip('`')

        # remove `foo`
        return content.strip('` \n')

    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @commands.command()
    async def execute(self, ctx: commands.Context, *, content: str):
        """
        Evaluates some python code
        Gracefully stolen from Rapptz ->
        https://github.com/Rapptz/RoboDanny/blob/rewrite/cogs/admin.py#L72-L117
        """

        # make the environment
        # this allows you to reference particular
        # variables in your block of code in discord
        # if you want to get technical, this is an
        # implementation of a Symbol Table
        env = {
            'bot': self.bot,
            'ctx': ctx,
            'channel': ctx.channel,
            'author': ctx.author,
            'guild': ctx.guild,
            'message': ctx.message,
            'self': self
        }
        env.update(globals())

        # make code and output string
        content = self._cleanup_code(content)
        stdout = StringIO()
        to_compile = f'async def func():\n{indent(content, "  ")}'

        # create the function to be ran
        try:
            exec(to_compile, env)
        except Exception as e:
            # compilation error
            # send the error in a code block
            return await ctx.send(f'```py\n{e.__class__.__name__}: {e}\n```')

        # execute the function we just made
        func = env["func"]
        try:
            # force stdout into StringIO
            with redirect_stdout(stdout):
                # run our function
                ret = await func()
        except Exception:
            # runtime error
            # output the error in a code block
            value = stdout.getvalue()
            await ctx.send(f'```py\n{value}{format_exc()}\n```')
        else:
            # hey we didn't goof up

            # react with a white check mark to show that it ran
            try:
                await ctx.message.add_reaction("\u2705")
            except Exception:
                pass

            # if nothing was returned, it might have printed something
            value = stdout.getvalue()
            if ret is None:
                if value:
                    await ctx.send(f'```py\n{value}\n```')
            else:
                text = f'```py\n{value}{ret}\n```'

                # character limits are a thing
                if len(text) > 2000:
                    # over max limit
                    # send output in a file
                    return await ctx.send(
                        file=discord.File(StringIO('\n'.join(text.split('\n')[1:-1])),
                                          filename='ev.txt')
                    )
                await ctx.send(text)
    
    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @commands.group(name="cog")
    async def _cog_group(self, ctx: commands.Context):
        """
        Cog related commands and utilities
        """

        pass

    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @_cog_group.command(name="list")
    async def _cog_list_command(self, ctx: commands.Context):
        """
        Lists all cogs currently running on the bot
        """

        await self.bot.paginate(
            ctx,
            "Cog List",
            "`{0}`",
            list(self.bot.cogs.keys()),
            max_page_size=1024,
            sequence_type_name="cogs",
            author_name=str(self.bot.user),
            author_icon_url=self.bot.user.avatar_url
        )

    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @_cog_group.command(name="info")
    async def _cog_info_command(self, ctx: commands.Context, cog: CogConverter):
        """
        Returns info about a given cog
        """

        embed = discord.Embed(
            title=f"Cog Info - {cog.qualified_name}",
            description=cog.description if cog.description != "" else "(No Description Provided)",
            color=self.bot.color,
            timestamp=datetime.datetime.utcnow()
        )

        embed.set_author(
            name=str(ctx.bot.user),
            icon_url=ctx.bot.user.avatar_url
        )

        listeners = cog.get_listeners()
        fmt = ', '.join(sorted([f'`{listener[0]}`' for listener in listeners])) if len(listeners) != 0 else "None"

        embed.add_field(
            name=f"Listeners[{len(listeners)}]",
            value=fmt,
            inline=False
        )

        commands = cog.get_commands()
        fmt = ', '.join(sorted([f'`{cmd.name}`' for cmd in commands])) if len(commands) != 0 else "None"
        embed.add_field(
            name=f"Commands[{len(commands)}]",
            value=fmt
        )

        await ctx.send(embed=embed)
    
    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @_cog_group.command(name="load")
    async def _cog_load_command(self, ctx: commands.Context, cog: str):
        """
        Loads a particular cog
        """

        self.bot.load_extension(cog)
        await self.bot.send_success_embed(ctx, f"Cog `{cog}` was loaded!")
    
    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @_cog_group.command(name="unload")
    async def _cog_unload_command(self, ctx: commands.Context, cog: str):
        """
        Unloads a particular cog
        """

        self.bot.unload_extension(cog)
        await self.bot.send_success_embed(ctx, f"Cog `{cog}` was unloaded!")
    
    @commands.is_owner()
    @commands.bot_has_permissions(send_messages=True)
    @_cog_group.command(name="reload")
    async def _cog_reload_command(self, ctx: commands.Context, cog: str):
        self.bot.unload_extension(cog)
        self.bot.load_extension(cog)
        await self.bot.send_success_embed(ctx, f"Cog `{cog}` was reloaded!")

    @_cog_info_command.error
    async def on_cog_commands_error(self, ctx: commands.Context, error):
        if isinstance(error, commands.ExtensionNotFound):
            await self.bot.send_fail_embed(ctx, f"Cog `{error.name}` was not found")
        elif isinstance(error, commands.ExtensionAlreadyLoaded):
            await self.bot.send_fail_embed(ctx, f"Cog `{error.name}` is already loaded")
        elif isinstance(error, commands.ExtensionNotLoaded):
            await self.bot.send_fail_embed(ctx, f"Cog `{error.name}` is already unloaded")
        elif isinstance(error, commands.NoEntryPointError):
            await self.bot.send_fail_embed(ctx, f"Cog `{error.name}` is missing `setup` entrypoint")
        elif isinstance(error, commands.ExtensionFailed):
            await self.bot.send_fail_embed(ctx, f"Cog {error.name} initiation failed: `{error.original.__class__.__name__}`")
        
        print(error, error.__class__.__name__)

    @_cog_load_command.error
    async def on_cog_load_command_error(self, ctx: commands.Context, error):
        await self.on_cog_commands_error(ctx, error)
    
    @_cog_unload_command.error
    async def on_cog_unload_command_error(self, ctx: commands.Context, error):
        await self.on_cog_commands_error(ctx, error)



import discord
import datetime
from discord.ext import commands
from typing import Any, Iterator, Optional, Sequence, Union
from .constants import MAX_EMBED_TITLE_LENGTH, MAX_EMBED_DESCRIPTION_LENGTH
from .utils import get_discord_color, get_base_embed


class Page:
    """Represents a single page in a paginator

    This class allows for a little lower level of control in terms of how each page in a
    paginator is displayed

    Parameters
    ----------
    title : str
        The title for the page, shown in the page's embed title
    items : Sequence[Any]
        The sequence to iterate over for the page
    line_sep : Optional str
        The line separator to join for each line.
        Defaults to "\\n"
    prefix : Optional str
        The prefix to prepend at the beginning of the page's embed description
        Defaults to "\`\`\`"
    suffix : Optional str
        The suffix to append at the end of the page's embed description.
        Defaults to "\`\`\`"
    color : Optional Union(discord.Color, tuple, str)
        The color to use for the Discord Embed. Can be either a `discord.Color` object,
        a hexadecimal color string, or an RGB tuple.
        Defaults to `discord.Color.dark_theme()`
    
    Attributes
    ----------
    embed : discord.Embed
        The page's Discord Embed that will be sent in Discord for pagination
    """

    def __init__(
        self,
        title : str,
        items: Sequence[Any],
        line_sep: Optional[str] = "\n",
        prefix: Optional[str] = "```",
        suffix: Optional[str] = "```",
        color: Optional[Union[discord.Color, tuple, str]] = discord.Color.dark_theme(),
        timestamp: Optional[datetime.datetime] = datetime.datetime.utcnow()
    ):
        # TODO: Add parameters for embed author, description, thumbnail, and image customization
        
        self.title = title
        self.description = line_sep.join([prefix, *[item for item in items], suffix])
        self.color = get_discord_color(color)
        self.timestamp = timestamp

    def __str__(self) -> str:
        """Returns the string to be used for this `Page`'s embed description

        Returns
        -------
        This `Page`'s embed description
        """

        return self.embed.description 

    @property
    def embed(self) -> discord.Embed:
        """Returns the Discord Embed representation for this paginator page

        Returns
        -------
        A `discord.Embed` for the page
        """

        return get_base_embed(
            title=self.title,
            description=self.description,
            color=self.color,
            timestamp=self.timestamp
        )


class Paginator:
    def __init__(self):
        self.pages = []

    def insert_page(self, index: int, page: Page):
        """Inserts a new page at a particular position in the paginator

        Prameters
        ---------
        page : Page
            The page to insert into the paginator
        """

        self.pages.insert(index, page)

    def prepend_page(self, page: Page):
        """Adds a new page to the beginning of the paginator's pages

        Prameters
        ---------
        page : Page
            The page to add to the beginning of the paginator
        """

        self.pages.insert(0, page)

    def add_page(self, page: Page):
        """Adds a new page to the end of the paginator's pages

        NOTE: Using this method assumes that you will be creating each page manually,
        ie. not via `Paginator.set_sequence`.
        """

        self.pages.append(page)

    def __iter__(self) -> Iterator[Page]:
        """Returns an interator to iterate through the paginator's pages

        Returns
        -------
        An iterator of `Page`s
        """

        return iter(self.pages)
    
    def __next__(self) -> Optional[Page]:
        """Returns the next page in the paginator

        Raises
        ------
        `StopIteration` when there are no more pages to paginate through

        Returns
        -------
        The next `Page`
        """

        return next(self.pages)

    def __len__(self) -> int:
        """Returns the number of pages for the paginator
        
        Returns
        -------
        An integer representing the number of pages the paginator has
        """

        return len(self.pages)

    @property
    def is_paginated(self) -> bool:
        """Returns whether the paginator is "paginated", meaning containing more than one page

        Returns
        -------
        `True` if the paginator contains more than one page, else `False`
        """

        return len(self) != 0

    def set_sequence(self, sequence: Sequence[Any]):
        """Sets the sequence to paginate

        TODO: This currently assumes that all sequences are two-dimensioned sequences

        NOTE: This overrrides all of the internal pages set. This means that even if you manually created your own pages
        using `Paginator.add_page()`, these pages will be overwritten.

        Parameters
        ----------
        sequence : Sequence(Any)
            The sequence to paginate
        """

        self.pages = [Page(item) for item in sequence]

    @commands.bot_has_permissions(send_messages=True, embed_links=True, add_reactions=True, manage_messages=True)
    async def paginate(self, ctx: commands.Context):
        """Starts the paginator in the given context

        NOTE: In order to paginate, your bot needs to have the
        following permissions in the given context:
        - Send Messages
        - Embed Links
        - Add Reactions
        - Manage Messages (for resetting pagination menu button reactions)
        """

        # set emojis
        far_left = "⏮"
        left = '⏪'
        right = '⏩'
        far_right = "⏭"

        # reaction check to be used later
        def predicate(m: discord.Message, set_begin: bool, push_left: bool, push_right: bool, set_end: bool):
            def check(reaction: discord.Reaction, user: discord.User):
                if reaction.message.id != m.id or user.id == ctx.bot.user.id or user.id != ctx.author.id:
                    return False
                if set_begin and reaction.emoji == far_left:
                    return True
                if push_left and reaction.emoji == left:
                    return True
                if push_right and reaction.emoji == right:
                    return True
                if set_end and reaction.emoji == far_right:
                    return True

                return False

            return check

        index = 0
        message = None
        action = ctx.send
        while True:
            res = await action(embed=self.pages[index].embed)

            if res is not None:
                message = res

            await message.clear_reactions()

            # determine which emojis should be added depending on how many pages are left in each direction
            set_begin = index > 1
            push_left = index != 0
            push_right = index != len(self.pages) - 1
            set_end = index < len(self.pages) - 2

            # add the appropriate emojis
            if set_begin:
                await message.add_reaction(far_left)
            if push_left:
                await message.add_reaction(left)
            if push_right:
                await message.add_reaction(right)
            if set_end:
                await message.add_reaction(far_right)

            # wait for reaction and set page index
            react, usr = await ctx.bot.wait_for(
                "reaction_add", check=predicate(message, set_begin, push_left, push_right, set_end)
            )

            # set next page index
            if react.emoji == far_left:
                index = 0
            elif react.emoji == left:
                index -= 1
            elif react.emoji == right:
                index += 1
            elif react.emoji == far_right:
                index = len(self.pages) - 1
            else:
                # invalid reaction, remove it
                await react.remove(usr)

            action = message.edit

    @classmethod
    def from_pages(cls, *pages: Sequence[Page]):
        """Creates a paginator from a given list of pages

        This allows for more lower level control than `Paginator.from_sequence`

        Parameters
        ----------
        pages : Sequence[Page]
            The list of pages to add to the list
        """

        c = cls()
        c.pages = pages
        return c

    @classmethod
    def from_sequence(cls, sequence: Sequence[Any]):
        """Creates a default paginator from a given sequence

        This mainly serves as a shortcut to creating a default `Paginator` object
        then setting the sequence seperately.

        Parameters
        ----------
        sequence : Sequence(Any)
            The sequence to create the paginator from
        
        Returns
        -------
        A `Paginator` with its sequence set to the given sequence
        """

        return cls().set_sequence(sequence)



MAX_EMBED_TITLE_LENGTH = 256
MAX_EMBED_DESCRIPTION_LENGTH = 2048
MAX_EMBED_FOOTER_TEXT_LENGTH = 2048
MAX_EMBED_AUTHOR_NAME_LENGTH = 256
MAX_EMBED_TOTAL_LENGTH = 6000
MAX_EMBED_FIELD_NAME_LENGTH = 256
MAX_EMBED_FIELD_VALUE_LENGTH = 1024
MAX_EMBED_FIELD_COUNT = 25


import aiosqlite
from typing import Any, Iterable, List, Optional, Union
from enum import Enum


"""
PRAGMA foreign_keys = 0;

CREATE TABLE cheesyutils_temp_table AS SELECT * FROM {};

DROP TABLE {};

CREATE TABLE blacklist (
    server_id INTEGER NOT NULL,
    username  TEXT    DEFAULT NULL,
    user_id   INTEGER NOT NULL,
    mod_id    INTEGER NOT NULL,
    timestamp DOUBLE  NOT NULL,
    reason    TEXT    DEFAULT NULL,
    test      CHAR
);

INSERT INTO {} (
                          server_id,
                          username,
                          user_id,
                          mod_id,
                          timestamp,
                          reason
                      )
                      SELECT server_id,
                             username,
                             user_id,
                             mod_id,
                             timestamp,
                             reason
                        FROM cheesyutils_temp_table;

DROP TABLE cheesyutils_temp_table;

PRAGMA foreign_keys = 1;
"""


class DataType(Enum):
    integer = "INTEGER"
    real = "REAL"
    blob = "BLOB"
    text = "TEXT"
    null = "NULL"


class Table:
    """Helper class to construct sql queries for tables"""

    def __init__(self, name: str):
        self._name = name
        self._columns = []
        self._has_primary_key = False

    @property
    def sql(self) -> Optional[str]:    
        return """CREATE TABLE IF NOT EXISTS {} {}""".format(self._name, "(" + (", ".join([column for column in self._columns])) + ")")
    
    def add_column(
        self,
        name: str,
        datatype: DataType,
        primary_key: Optional[bool] = False,
        unique: Optional[bool] = False,
        default: Optional[Any] = None,
        not_null: Optional[bool] = False
    ):
        sql = f"""{name} {datatype.value}"""

        if primary_key:
            if self._has_primary_key:
                raise ValueError("Primary key already exists in table")
            else:
                sql += """ PRIMARY KEY"""

        if not_null:
            sql += """ NOT NULL"""
        
        if unique:
            sql += """ UNIQUE"""
        
        if default:
            # add redundant quotes on strings
            if isinstance(default, str):
                default = f"\"{default}\""

            sql += f""" DEFAULT {default}"""


        self._columns.append(sql)
    

class Database:
    def __init__(self, fp: str):
        self.fp = fp

    async def create_table(self, table: Table):
        await self.execute(table.sql)

    async def query_first(self, sql: str, as_dict: Optional[bool] = True, params: Iterable[Any] = None) -> Optional[Union[dict, list]]:
        res = await self.execute(sql, auto_commit=False, as_dict=as_dict, params=params)
        if isinstance(res, (list, tuple)):
            return res[0] if res else None
        return res
    
    async def query_all(self, sql: str, as_dict: Optional[bool] = True, params: Iterable[Any] = None) -> Optional[Union[list, dict]]:
        return await self.execute(sql, auto_commit=False, as_dict=as_dict, params=params)

    async def execute(self, sql: str, auto_commit: Optional[bool] = True, as_dict: Optional[bool] = False, params: Iterable[Any] = None) -> List[dict]:
        async with aiosqlite.connect(self.fp) as db:
            if as_dict:
                db.row_factory = aiosqlite.Row

            async with db.execute_fetchall(sql, params=params) as result:
                if auto_commit:
                    await db.commit()

                return result


if __name__ == "__main__":
    table = Table("test_table")
    table.add_column(name="id", datatype=DataType.integer, not_null=True, primary_key=True)
    table.add_column(name="test", datatype=DataType.real, default=0.0)
    print(table.sql)



import datetime
import discord
from discord.ext import commands
from typing import List, Optional, Union
from .constants import *


def get_discord_color(color: Union[discord.Color, tuple, str]) -> discord.Color:
    """Returns a discord.Color from an RGB tuple or hex string
    
    The hex string parsing is case insensitive

    Parameters
    ----------
    color: Union(discord.Color, tuple, str)

    Raises
    ------
    TypeError if the color was not a discord.Color, tuple, or string

    Returns
    -------
    A discord.Color object
    """

    if type(color) is tuple:
        # assuming it's RGB
        return discord.Color.from_rgb(color[0], color[1], color[2])
    elif type(color) is str:
        # code snippet taken from https://stackoverflow.com/a/29643643
        return get_discord_color(tuple(int(color.lstrip("#")[i:i + 2], 16) for i in (0, 2, 4)))
    elif type(color) is discord.Color:
        return color
    else:
        raise TypeError("Invalid Color type. Must be discord.Color, RGB tuple, or hex string")


def get_image_url(entity: Union[discord.abc.User, discord.Guild, discord.Asset, str]):
    """Returns an image url depending on if the desired entity is a User, Guild, or string

    Parameters
    ----------
    entity : Union(discord.abc.User, discord.Guild, discord.Asset, str)
        The entity to get the image url from
    
    Returns
    -------
    The image url as a string, or the entity itself if its already a string
    """

    if isinstance(entity, (discord.abc.User, discord.Guild, discord.Asset, str)):
        if isinstance(entity, discord.abc.User):
            return str(entity.avatar_url)
        elif isinstance(entity, discord.Guild):
            return str(entity.icon_url)
        elif isinstance(entity, discord.Asset):
            return str(entity)

        return entity
    else:
        raise TypeError(f"Expected discord.abc.User, discord.Guild, or string, got \"{entity.__class__.__name__}\" instead")


def truncate(string: str, max_length: int, end: Optional[str] = "...") -> str:
    """Truncates a string

    Parameters
    ----------
    string : str
        The string to truncate, if needed
    max_length : int
        The maximum length of the string before truncation is needed
    end : Optional str
        The string to append to the end of the string after truncation
        The string is automatically downsized to accommodate the size of `end`
        This automatically defaults to "..."
    
    Raises
    ------
    ValueError
        If the size of `end` is larger than `max_length`

    Returns
    -------
    The truncated string
    """

    if len(end) > max_length:
        raise ValueError(f"End string \"{end}\" of length {len(end)} can't be larger than {max_length} characters")
    
    truncated = string[:max_length]
    if string != truncated:
        truncated = string[:max_length - len(end)] + end

    return truncated


def slice_string(string: str, max_length: int) -> List[str]:
    """Returns a list of strings of a particular maximum length
    
    Original solution taken from https://stackoverflow.com/a/18854817

    Parameters
    ----------
    string : str
        The string to slice
    max_length : int
        The maximum length for each string slice
    
    Returns
    -------
    A list of strings with maximum length of `max_length`
    """

    return [string[0+i:max_length+i] for i in range(0, len(string), max_length)]


def get_base_embed(
    title: Optional[str] = discord.Embed.Empty,
    description: Optional[str] = discord.Embed.Empty,
    color: Optional[Union[discord.Color, tuple, str]] = discord.Embed.Empty,
    timestamp: Optional[datetime.datetime] = datetime.datetime.utcnow(),
    author: Optional[Union[discord.abc.User, discord.Guild]] = None,
    author_name: Optional[str] = None,
    author_icon: Optional[Union[discord.abc.User, discord.Guild, str]] = discord.Embed.Empty,
    author_url: Optional[str] = discord.Embed.Empty,
    thumbnail: Optional[Union[discord.abc.User, discord.Guild, str]] = None,
    image: Optional[Union[discord.abc.User, discord.Guild, str]] = None,
    footer: Optional[Union[discord.abc.User, discord.Guild]] = None,
    footer_text: Optional[str] = discord.Embed.Empty,
    footer_icon: Optional[Union[discord.abc.User, discord.Guild, str]] = discord.Embed.Empty,
    url: Optional[str] = discord.Embed.Empty,
) -> discord.Embed:
    """Returns a "base" embed without any fields

    This function automatically truncates any excess characters for each field with "..."

    Parameters
    ----------
    title : Optional str
        The embed title
    description : Optional str
        The embed description
    color : Optional Union(discord.Color, tuple, str)
        The color to use for the embed. Can be a `discord.Color` object, RGB tuple, or hex string
    timestamp : Optional datetime.datetime
        The utc timestamp to use for the embed (defaults to the current utc timestamp)
    author : Optional Union(discord.abc.User, discord.Guild)
        The user or guild to use for the embed author name/icon. This overrides both `author_name` and `author_icon`
    author_name : Optional str
        The string to use for the embed author name
    author_icon : Optional str
        The url to use for the embed author icon
    author_url : Optional str
        The url to use for the embed author (for when the author name is clicked)
    thumbnail : Optional Union(discord.abc.User, discord.Guild, str)
        The url of the image to use for the embed thumbnail
    image : Optional Union(discord.abc.User, discord.Guild, str)
        The url of the image to use for the embed image
    footer : Optional Union(discord.abc.User, discord.Guild)
        The user or guild to use for the embed footer text/icon. This overrides both `footer_text` and `footer_icon`
    footer_text : Optional str
        The string to use for the footer text
    footer_icon : Optional str
        The url to use for the embed footer icon
    url : Optional str
        The url to use for the embed (for when the title is clicked)

    Raises
    ------
    ValueError
        - If the total embed character limit of 6000 characters is exceeded after attempting to truncate the embed title,
        description, footer text, and author_name
        - If all fields are empty
    
    Returns
    -------
    A discord.Embed from the provided attributes, without any embed fields
    """

    Empty = discord.Embed.Empty

    # check if all None
    if title is Empty and description is Empty and color is Empty and author is None and author_name is None and thumbnail is None and image is None and footer is None and footer_text is Empty and footer_icon is Empty and url is None:
        raise ValueError("Cannot construct an empty base embed")

    # set author icon
    if author_icon is not Empty:
        author_icon = get_image_url(author_icon)

    # override author attributes
    if author is not None:
        author_name = str(author)
        author_icon = author.avatar_url if isinstance(author, discord.abc.User) else author.icon_url
    
    # set footer icon
    if footer_icon is not Empty:
        footer_icon = get_image_url(footer_icon)

    # override footer attributes
    if footer is not None:
        footer_text = str(footer)
        footer_icon = footer.avatar_url if isinstance(footer, discord.abc.User) else footer.icon_url

    # truncate neccessary fields
    title = truncate(title, MAX_EMBED_TITLE_LENGTH) if title is not discord.Embed.Empty else discord.Embed.Empty
    description = truncate(description, MAX_EMBED_DESCRIPTION_LENGTH) if description is not discord.Embed.Empty else discord.Embed.Empty
    footer_text = truncate(footer_text, MAX_EMBED_FOOTER_TEXT_LENGTH) if footer_text is not Empty else Empty
    author_name = truncate(author_name, MAX_EMBED_AUTHOR_NAME_LENGTH) if author_name is not None else None

    # construct embed
    embed = discord.Embed(
        title=title,
        description=description,
        color=get_discord_color(color),
        timestamp=timestamp,
        url=url,
    )

    # set embed author
    if author_name is not None:
        embed.set_author(
            name=author_name,
            icon_url=author_icon,
            url=author_url
        )
    
    # set embed footer
    if footer_text is not Empty and footer_icon is not Empty:
        embed.set_footer(
            text=footer_text,
            icon_url=footer_icon
        )
    
    # set embed thumbnail
    if thumbnail is not None:
        embed.set_thumbnail(get_image_url(thumbnail))
    
    # set embed image
    if image is not None:
        embed.set_image(get_image_url(image))
    
    # check max length
    if len(embed) > MAX_EMBED_TOTAL_LENGTH:
        raise ValueError(f"Base embed character count of {len(embed)} is over limit of {MAX_EMBED_TOTAL_LENGTH}")
    
    return embed


async def paginate(
        ctx: commands.Context,
        embed_title: str,
        line: str,
        sequence: list,
        embed_color: Optional[discord.Color] = discord.Color.dark_theme(),
        prefix: Optional[str] = "",
        suffix: Optional[str] = "",
        max_page_size: Optional[int] = 2048,
        other_sequence: Optional[list] = None,
        sequence_type_name: Optional[str] = None,
        author_name: Optional[str] = None,
        author_icon_url: Optional[str] = None,
        count_format: Optional[str] = None
):
    """Creates a paginating menu given a particular context

    Warning: The following permissions are required for the bot to utilize pagination
    - Add Reactions
    - Manage Messages

    Parameters
    ----------
    ctx : commands.Context
        The context in which the pagination is to take place
    embed_title : str
        The title to use for the embed
    line : str
        The format string to use for each line in a page. Each item in `sequence` gets its own line in a page
    sequence : list
        The sequence in which pagination should take place
    embed_color : Optional discord.Color
        The embed color to use. Defaults to `discord.Color.dark_theme()`
    prefix : Optional str
        The string to prepend to the beginning of the page (defaults to an empty string)
    suffix : Optional str
        The string to append to the end of the page (defaults to an empty string)
    max_page_size : Optional int
        The maximum number of characters per page (defaults to 2048)
    other_sequence : Optional list
        The other sequence to iterate over (mainly for the embed's footer text, defaults to be equal to `sequence`)
    sequence_type_name : Optional str
        The name of the other sequence (mainly for the embed's footer text, defaults to "items")
    author_name : Optional str
        The name to use for the embed author (defaults to the current guild's name)
    author_icon_url : Optional str
        The icon url to use for the embed author (defaults to the current guild's icon)
    count_format : Optional str
        The string to prepend to each line (used for a count)
    """

    if other_sequence is None:
        other_sequence = sequence
        sequence_type_name = "items" if sequence_type_name is None else sequence_type_name

    far_left = "⏮"
    left = '⏪'
    right = '⏩'
    far_right = "⏭"

    def predicate(m: discord.Message, set_begin: bool, push_left: bool, push_right: bool, set_end: bool):
        def check(reaction: discord.Reaction, user: discord.User):
            if reaction.message.id != m.id or user.id == ctx.bot.user.id or user.id != ctx.author.id:
                return False
            if set_begin and reaction.emoji == far_left:
                return True
            if push_left and reaction.emoji == left:
                return True
            if push_right and reaction.emoji == right:
                return True
            if set_end and reaction.emoji == far_right:
                return True

            return False

        return check

    # init paginator
    paginator = commands.Paginator(prefix=prefix, suffix=suffix, max_size=max_page_size)

    item_count = 0
    for item in sequence:
        item_count += 1
        if count_format is not None:
            paginator.add_line(
                line=(count_format.format(item_count) + line.format(item))
            )
        else:
            paginator.add_line(
                line=line.format(item)
            )

    index = 0
    message = None
    action = ctx.send
    while True:
        # this is probably going to fuck shit up
        embed = discord.Embed(
            title=embed_title,
            description=paginator.pages[index],
            color=embed_color,
            timestamp=datetime.datetime.utcnow()
        )

        if author_name is None:
            author_name = ctx.guild.name
        
        if author_icon_url is None:
            author_icon_url = ctx.guild.icon_url

        embed.set_author(
            name=author_name,
            icon_url=author_icon_url
        )

        embed.set_footer(
            text=f"Page {index + 1}/{len(paginator.pages)} • "
                    f"{len(sequence)}/{len(other_sequence)} {sequence_type_name}"
        )

        res = await action(embed=embed)

        if res is not None:
            message = res

        await message.clear_reactions()

        # determine which emojis should be added
        set_begin = index > 1
        push_left = index != 0
        push_right = index != len(paginator.pages) - 1
        set_end = index < len(paginator.pages) - 2

        # add the appropriate emojis
        if set_begin:
            await message.add_reaction(far_left)
        if push_left:
            await message.add_reaction(left)
        if push_right:
            await message.add_reaction(right)
        if set_end:
            await message.add_reaction(far_right)

        # wait for reaction and set page index
        react, usr = await ctx.bot.wait_for(
            "reaction_add", check=predicate(message, set_begin, push_left, push_right, set_end)
        )

        if react.emoji == far_left:
            index = 0
        elif react.emoji == left:
            index -= 1
        elif react.emoji == right:
            index += 1
        elif react.emoji == far_right:
            index = len(paginator.pages) - 1

        action = message.edit


import aiohttp
import aiosqlite
import asyncio
import discord
import os
import re
from discord.ext import commands
from typing import Any, Awaitable, Callable, Optional, Union
from .help_command import HelpCommand
from .cogs.meta import Meta
from .utils import *
from .database import *


class DiscordBot(commands.Bot):
    """
    Defines a Discord Bot
    """

    def __init__(
            self,
            prefix: Optional[str] = ".",
            color: Optional[Union[discord.Color, tuple, str]] = discord.Color.dark_theme(),
            database: Optional[str] = "bot.db",
            members_intent: Optional[bool] = False,
            presences_intent: Optional[bool] = False,
            status: Optional[str] = "online",
            activity: Optional[str] = None,
            loop: Optional[asyncio.BaseEventLoop] = None,
            custom_emojis: Optional[dict] = None,
            *args, **kwargs
    ):
        """Defines a new Discord Bot

        A lot of these attributes are optional and can be completely omitted if you so wish

        Attributes
        ----------
        prefix : Optional str
            The command prefix for the bot (defaults to ".")
        color : Optional Union(discord.Color, tuple, str)
            The default embed color to use for sending Discord embeds (defaults to `discord.Color.dark_theme()`)
            This parameter could be a `discord.Color` object, a hexadecimal color code, or an RGB tuple. The color will always
            end up being converted to a `discord.Color` object
        database : Optional str
            The name of the bot's sqlite database file (defaults to "bot.db")
        members_intent : Optional bool
            Whether to requests the `members` Discord API Intent (defaults to False).
            If this is set to true, you will need to ensure your bot has the
            `members` privileged intent enabled in the Discord Developer Portal
        presences_intent : Optional bool
            Whether to request the `presences` Discord API Intent (defaults to False)
            If this is set to true, you will need to ensure your bot has the
            `presences` privileged intent enabled in the Discord Developer Portal
        status : Optional str
            The online status to set for the bot (defaults to "online").
            This is case-insensitive and can be "online", "idle", "afk", "dnd", or "invisible"
        activity : Optional str
            The "playing" activity to set for the bot (defaults to "with Cheesy | (PREFIX)help)
        loop : Optional asyncio.BaseEventLoop
            The asyncio event loop to set for the bot (defaults to the value returned by `asyncio.get_event_loop()`)
        custom_emojis : Optional dict
            A dictionary who's string key corresponds to a particular Discord emoji string. Used for `Discord.Bot.get_emoji`
        """

        self.loop = loop or asyncio.get_event_loop()
        self.color = get_discord_color(color)

        # set intents
        intents = discord.Intents.default()
        intents.members = members_intent
        intents.presences = presences_intent

        # set activity
        if activity is None:
            activity = f"with Cheesy | {prefix}help"

        # set database
        self.database = Database(database)

        self.is_actually_ready = False

        super().__init__(
            command_prefix=commands.when_mentioned_or(prefix),
            intents=intents,
            activity=discord.Game(name=activity),
            status=self.get_discord_status(status),
            help_command=HelpCommand(self.color),
            *args, **kwargs
        )

        self.custom_emojis = custom_emojis

        self.add_cog(Meta(self))

    async def on_ready(self):
        if not self.is_actually_ready:
            # get current server id's
            #server_ids = await self.database.query_all(Table("config").sql)
            self.is_actually_ready = True
            print(f"{self.user} is ready!")
    
    def run(self, token: Union[os.PathLike, str]):
        """Starts the bot

        Starting the bot can be done by the following methods:
        - Providing the raw token
        - Providing the path to a text file containing the token.
        This is done by reading the first line of the file and stripping trailing spaces

        Parameters
        ----------
        token : Union(os.PathLike, str)
            The token or the path to a text file containing the token
        """

        try:
            with open(token, "r") as file:
                token = file.readlines()[0].strip()
        except (FileNotFoundError, PermissionError, FileExistsError) as e:
            print(f"Assuming raw token passed - could not find token file due to an error: \"{e.__class__.__name__}\"")
            token = token

        try:
            super().run(token)
        except (aiohttp.ClientConnectorError, discord.HTTPException, discord.LoginFailure) as e:
            print(f"Unable to start bot: \"{type(e)}\" - \"{e}\"")

    def get_emoji(self, id: Union[str, int]) -> Optional[discord.Emoji]:
        """Returns a Discord Emoji

        This method slightly differs from discord.py's implementation.
        For starters, the bot will first look inside it's user-defined emoji reference (see `DiscordBot.custom_emojis`).
        From there, the bot will continue to search for the emoji in it's cache similar to discord.py's implementation

        Parameters
        ----------
        id : Union(str, int)
            The Discord ID OR a key for `DiscordBot.custom_emojis`
        
        Returns
        -------
        Optional discord.Emoji
        """

        if isinstance(id, str):
            # lookup key in the bot's custom emoji key
            if self.custom_emojis is not None:
                emoji_str = self.custom_emojis.get(id)
                
                if emoji_str is not None:
                    def extract_emoji_id(emoji: str) -> int:
                        return int(re.match(r'<a?:[a-zA-Z0-9\_]+:([0-9]+)>$', emoji).group(1))

                    return super().get_emoji(extract_emoji_id(emoji_str))

        # defaults to parent method should the above fail
        return super().get_emoji(id)

    async def paginate(
        self,
        ctx: commands.Context,
        embed_title: str,
        line: str,
        sequence: list,
        prefix: Optional[str] = "",
        suffix: Optional[str] = "",
        max_page_size: Optional[int] = 2048,
        other_sequence: Optional[list] = None,
        sequence_type_name: Optional[str] = None,
        author_name: Optional[str] = None,
        author_icon_url: Optional[str] = None,
        count_format: Optional[str] = None
    ):
        """Creates a paginating menu given a particular context

        Warning: The following permissions are required for the bot to utilize pagination
        - Add Reactions
        - Manage Messages

        Parameters
        ----------
        ctx : commands.Context
            The context in which the pagination is to take place
        embed_title : str
            The title to use for the embed
        line : str
            The format string to use for each line in a page. Each item in `sequence` gets its own line in a page
        sequence : list
            The sequence in which pagination should take place
        embed_color : Optional discord.Color
            The embed color to use. Defaults to `discord.Color.dark_theme()`
        prefix : Optional str
            The string to prepend to the beginning of the page (defaults to an empty string)
        suffix : Optional str
            The string to append to the end of the page (defaults to an empty string)
        max_page_size : Optional int
            The maximum number of characters per page (defaults to 2048)
        other_sequence : Optional list
            The other sequence to iterate over (mainly for the embed's footer text, defaults to be equal to `sequence`)
        sequence_type_name : Optional str
            The name of the other sequence (mainly for the embed's footer text, defaults to "items")
        author_name : Optional str
            The name to use for the embed author (defaults to the current guild's name)
        author_icon_url : Optional str
            The icon url to use for the embed author (defaults to the current guild's icon)
        count_format : Optional str
            The string to prepend to each line (used for a count)
        """

        await paginate(
            ctx, embed_title, line, sequence, self.color,
            prefix, suffix, max_page_size, other_sequence,
            sequence_type_name, author_name, author_icon_url, count_format
        )
    
    @staticmethod
    def get_discord_status(status: str) -> discord.Status:
        """Returns a discord status from a string

        This is mainly used for bot initiation

        The provided string is automatically converted to lowercase, providing the following valid options:
        - "online" (Online)
        - "away" (Idle)
        - "idle" (Idle)
        - "dnd" (Do not Disturb)
        - "do not disturb" (Do not Disturb)

        Parameters
        ----------
        status : str
            The string to convert into a discord.Status
        
        Raises
        ------
        ValueError if the provided string could not be converted to a discord.Status

        Returns
        -------
        A discord.Status representation of the given string
        """

        status_lower = status.lower()

        if status_lower == "dnd" or status_lower == "do not disturb":
            return discord.Status.dnd
        elif status_lower == "idle" or status_lower == "away":
            return discord.Status.idle
        elif status_lower == "online":
            return discord.Status.online
        else:
            raise ValueError(f"Invalid status string \"{status}\"")
    @commands.command
    @staticmethod
    async def _retrieve_entity(snowflake: int, func: Callable[[int], Any], coro: Awaitable[int]):
        """Lazily retrieves a discord object from a given discord id
        
        This function is moreso designed to be a helper function for other `retrieve_X` coroutines,
        which can be found below, in order to lazily fetch discord objects in order to reduce API calls

        Parameters
        ----------
        snowflake : int
            The discord id of the object to retrieve
        func : Callable(int)
            The function to initially call before attempting to fetch using `coro` if needed
        coro : Awaitable(int)
            The coroutine to await should the result of `func` be equal to None
        
        Returns
        -------
        The respective Discord object, or None if the object was unable to be retrieved
        """

        entity = func(snowflake)
        if entity is None:
            try:
                return await coro(snowflake)
            except (discord.HTTPException, discord.InvalidData):
                return None
        else:
            return entity

    async def retrieve_channel(
            self,
            channel_id: int
    ) -> Optional[Union[discord.abc.GuildChannel, discord.abc.PrivateChannel]]:
        """Lazily retrieves a Discord channel

        NOTE: Per discord.py's documentation, you can only retrieve channels
        that the bot currently has access to. This means that a given channel id
        could be valid, but if the bot doesn't have access to it, the bot will not
        be able to retrieve it. In that case, this function will return None

        Parameters
        ----------
        channel_id : int
            The Discord ID of the channel you wish to retrieve

        Returns
        -------
        An discord.abc.GuildChannel for guild-specific channels,
        or discord.abc.PrivateChannel for user specific channels (DMs),
        or None if the channel could not be retrieved
        """

        return await self._retrieve_entity(channel_id, self.get_channel, self.fetch_channel)

    async def retrieve_guild(self, guild_id: int) -> Optional[discord.Guild]:
        """Lazily retrieves a Discord guild

        NOTE: Per discord.py's documentation, you can only retrieve guilds
        that the bot currently has access to. This means that a given guild id
        could be valid, but if the bot doesn't have access to it, the bot will not
        be able to retrieve it. In that case, this function will return None

        Parameters
        ----------
        guild_id : int
            The Discord ID of the guild you wish to retrieve
        
        Returns
        -------
        A discord.Guild if the guild could be retrieved, or None of the guild
        was not found or if the bot does not have access to the respective guild
        """

        return await self._retrieve_entity(guild_id, self.get_guild, self.fetch_guild)

    async def retrieve_user(self, user_id: int) -> Optional[discord.User]:
        """Lazily retrieves a Discord User

        Parameters
        ----------
        user_id : int
            The Discord ID of the user you wish to retrieve
        
        Returns
        -------
        A discord.User if the user could be retrieved, None otherwise
        """

        return await self._retrieve_entity(user_id, self.get_user, self.fetch_user)

    async def retrieve_message(
            self,
            channel_id: Optional[int] = None,
            message_id: Optional[int] = None,
            message_link: Optional[str] = None
    ) -> Optional[discord.Message]:
        """Lazily retrieves a Discord Message

        There are two ways to retrieve messages via this method
        - By supplying the `channel_id` and `message_id` parameters with their respective values
        - By supplying the `message_link`
        - Supplying both of the above will retrieve from `message_id` and `channel_id` only
        
        NOTE: Per discord.py's documentation, you can only retrieve messages
        that the bot currently has access to. This means that a given message/channel id
        or message link could be valid, but if the bot doesn't have access to where the
        message is located, the bot will not be able to retrieve it. In that case, this
        function will return None

        Parameters
        ----------
        channel_id : Optional int
            The Discord ID of the channel where the message you wish to retrieve is located
        message_id : Optional int
            The Discord ID of the message you wish to retrieve
        message_link : Optional str
            The url of the message link for the message you wish to retrieve
            Note that if this parameter is supplied, this method is called again
            using the channel/message id extracted from the url

        Returns
        -------
        A discord.Message if the message could be retrieved, None otherwise
        """

        if channel_id is not None and message_id is not None:
            channel = await self.retrieve_channel(channel_id)
            if channel is not None:
                try:
                    return await channel.fetch_message(message_id)
                except discord.HTTPException:
                    return None
            else:
                return None
        elif message_link is not None:
            try:
                channel_id, message_id = message_link.split("/")[5:]
                return await self.retrieve_message(int(channel_id), int(message_id))
            except (IndexError, ValueError):
                return None
        else:
            return None

    async def retrieve_member(
            self,
            guild: Union[discord.Guild, int],
            user_id: int
    ) -> Optional[discord.Member]:
        """Lazily retrieves a Discord guild member

        NOTE: Per discord.py's documentation, you can only retrieve members from guilds
        that the bot currently has access to. This means that a given user/guild id
        could be valid, but if the bot doesn't have access to the guild, the bot will not
        be able to retrieve the member. In that case, this function will return None
        
        Parameters
        ----------
        guild : Union(discord.Guild, int)
            The guild id of the guild or the guild object itself to retrieve the desired member from.
            Note that if a guild id is supplied rather than a guild object, the guild will have to be
            retrieved
        user_id : int
            The Discord ID of the member you wish to retrieve

        Returns
        -------
        A discord.Member if the member could be retrieved, None otherwise
        """

        if isinstance(guild, discord.Guild):
            guild = guild
        elif isinstance(guild, int):
            guild = await self.retrieve_guild(guild)
            if guild is None:
                return None
        else:
            raise ValueError(f"Invalid Guild type. Must be discord.Guild or int, not \"{type(guild)}\"")

        member = guild.get_member(user_id)
        if member is None:
            try:
                return await guild.fetch_member(user_id)
            except discord.HTTPException:
                return None
        else:
            return member
